@@ -490,17 +490,24 @@ mkisofs_stdout_read (GIOChannel   *source,
 	struct mkisofs_output *mkisofs_output = data;
 	char *line;
 	GIOStatus status;
-	
-	status = g_io_channel_read_line (source,
-					 &line, NULL, NULL, NULL);
 
-	if (line && mkisofs_output->debug) {
-		g_print ("make_iso stdout: %s", line);
+	if (condition & G_IO_IN) {
+		status = g_io_channel_read_line (source,
+						 &line, NULL, NULL, NULL);
+
+		if (line && mkisofs_output->debug) {
+			g_print ("make_iso stdout: %s", line);
+		}
+
+		if (status == G_IO_STATUS_NORMAL) {
+			g_free (line);
+		}
+	} else if (condition & G_IO_HUP) {
+		if (mkisofs_output->debug)
+			g_print ("make_iso stdout: HUP\n");
+		return FALSE;
 	}
 
-	if (status == G_IO_STATUS_NORMAL) {
-		g_free (line);
-	}
 	return TRUE;
 }
 
@@ -515,61 +522,99 @@ mkisofs_stderr_read (GIOChannel   *source,
 	double                 fraction;
 	GIOStatus              status;
 
-	status = g_io_channel_read_line (source, &line, NULL, NULL, NULL);
+	if (condition & G_IO_IN) {
+		status = g_io_channel_read_line (source, &line, NULL, NULL, NULL);
 
-	if (line && mkisofs_output->debug) {
-		g_print ("make_iso stderr: %s", line);
-	}
-
-	if (status == G_IO_STATUS_NORMAL) {
-		if (strncmp (line, "Total translation table size", 28) == 0) {
-			nautilus_burn_progress_set_fraction (1.0);
-			g_main_loop_quit (mkisofs_output->loop);
-			mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_FINISHED;
+		if (line && mkisofs_output->debug) {
+			g_print ("make_iso stderr: %s", line);
 		}
 
-		if (strstr (line, "estimate finish")) {
-			if (sscanf (line, "%6c%% done, estimate finish",
-				    fraction_str) == 1) {
-				fraction_str [6] = 0;
-				fraction = g_strtod (fraction_str, NULL);
-				nautilus_burn_progress_set_fraction (fraction/100.0);
+		if (status == G_IO_STATUS_NORMAL) {
+			if (strncmp (line, "Total translation table size", 28) == 0) {
+				nautilus_burn_progress_set_fraction (1.0);
+				mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_FINISHED;
 			}
-		}
 
-		if (strstr (line, "Incorrectly encoded string")) {
-			nautilus_burn_progress_set_fraction (1.0);
-			g_main_loop_quit (mkisofs_output->loop);
-			mkisofs_output->error = g_error_new_literal (NAUTILUS_BURN_ISO_ERROR,
-								     NAUTILUS_BURN_ISO_ERROR_GENERAL,
-								     _("Incorrectly encoded string."));
-			mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
-		}
+			if (strstr (line, "estimate finish")) {
+				if (sscanf (line, "%6c%% done, estimate finish",
+					    fraction_str) == 1) {
+					fraction_str [6] = 0;
+					fraction = g_strtod (fraction_str, NULL);
+					nautilus_burn_progress_set_fraction (fraction/100.0);
+				}
+			}
 
-		if (strstr (line, "No space left on device")) {
-			nautilus_burn_progress_set_fraction (1.0);
-			g_main_loop_quit (mkisofs_output->loop);
-			mkisofs_output->error = g_error_new_literal (NAUTILUS_BURN_ISO_ERROR,
-								     NAUTILUS_BURN_ISO_ERROR_GENERAL,
-								     _("There is no space left on the device."));
-			mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
-		}
+			if (strstr (line, "Incorrectly encoded string")) {
+				nautilus_burn_progress_set_fraction (1.0);
+				if (!mkisofs_output->error)
+					mkisofs_output->error = g_error_new_literal (NAUTILUS_BURN_ISO_ERROR,
+										     NAUTILUS_BURN_ISO_ERROR_GENERAL,
+										     _("Incorrectly encoded string."));
+				mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+			}
 
-		if (strstr (line, "Value too large for defined data type")) {
-			/* TODO: get filename from error message */
-			nautilus_burn_progress_set_fraction (1.0);
-			g_main_loop_quit (mkisofs_output->loop);
-			mkisofs_output->error = g_error_new_literal (NAUTILUS_BURN_ISO_ERROR,
-								     NAUTILUS_BURN_ISO_ERROR_GENERAL,
-								     _("File too large for filesystem."));
-			mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
-		}
+			if (strstr (line, "Unknown charset")) {
+				nautilus_burn_progress_set_fraction (1.0);
+				/* FIXME: use better error message after string freeze */
+				if (!mkisofs_output->error)
+					mkisofs_output->error = g_error_new_literal (NAUTILUS_BURN_ISO_ERROR,
+										     NAUTILUS_BURN_ISO_ERROR_GENERAL,
+										     _("Incorrectly encoded string."));
+				mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+			}
 
-		g_free (line);
+			if (strstr (line, "No space left on device")) {
+				nautilus_burn_progress_set_fraction (1.0);
+				if (!mkisofs_output->error)
+					mkisofs_output->error = g_error_new_literal (NAUTILUS_BURN_ISO_ERROR,
+										     NAUTILUS_BURN_ISO_ERROR_GENERAL,
+										     _("There is no space left on the device."));
+				mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+			}
+
+			if (strstr (line, "Value too large for defined data type")) {
+				/* TODO: get filename from error message */
+				nautilus_burn_progress_set_fraction (1.0);
+				if (!mkisofs_output->error)
+					mkisofs_output->error = g_error_new_literal (NAUTILUS_BURN_ISO_ERROR,
+										     NAUTILUS_BURN_ISO_ERROR_GENERAL,
+										     _("File too large for filesystem."));
+				mkisofs_output->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+			}
+
+			g_free (line);
+		}
+	} else if (condition & G_IO_HUP) {
+		/* only handle the HUP when we have read all available lines of output */
+		if (mkisofs_output->debug)
+			g_print ("make_iso stderr: HUP\n");
+		nautilus_burn_progress_set_fraction (1.0);
+		g_main_loop_quit (mkisofs_output->loop);
+		return FALSE;
 	}
+
 	return TRUE;
 }
 
+/* this is an ugly hack until utf8/iconv support is added into upstream mkisofs */
+static gboolean
+ncb_mkisofs_supports_utf8 (void)
+{
+	char    *standard_error;
+	gboolean supported;
+	gboolean res;
+
+	res = g_spawn_command_line_sync ("mkisofs -input-charset utf8", NULL, &standard_error, NULL, NULL);
+	if (res && !g_strrstr (standard_error, "Unknown charset"))
+		supported = TRUE;
+	else
+		supported = FALSE;
+
+	g_free (standard_error);
+
+	return supported;
+}
+
 /**
  * nautilus_burn_make_iso:
  * @filename:
@@ -607,6 +652,7 @@ nautilus_burn_make_iso (const char *filename,
 	int                   exit_status, res;
 	guint64               iso_size;
 	GnomeVFSFileSize      size;
+	gboolean              has_utf8_support;
 
 	if (label && (strlen (label) > 32)) {
 		g_set_error (error,
@@ -665,6 +711,8 @@ nautilus_burn_make_iso (const char *filename,
 		goto cleanup;
 	}
 
+	has_utf8_support = ncb_mkisofs_supports_utf8 ();
+
  retry:
 	i = 0;
 	argv [i++] = "mkisofs";
@@ -672,9 +720,10 @@ nautilus_burn_make_iso (const char *filename,
 	if (use_joliet) {
 		argv [i++] = "-J";
 	}
-	/* Undocumented -input-charset option */
-	argv [i++] = "-input-charset";
-	argv [i++] = "utf8";
+	if (has_utf8_support) {
+		argv [i++] = "-input-charset";
+		argv [i++] = "utf8";
+	}
 	argv [i++] = "-q";
 	argv [i++] = "-graft-points";
 	argv [i++] = "-path-list";
@@ -758,8 +807,10 @@ nautilus_burn_make_iso (const char *filename,
 	if (use_joliet) {
 		argv [i++] = "-J";
 	}
-	argv [i++] = "-input-charset";
-	argv [i++] = "utf8";
+	if (has_utf8_support) {
+		argv [i++] = "-input-charset";
+		argv [i++] = "utf8";
+	}
 	argv [i++] = "-graft-points";
 	argv [i++] = "-path-list";
 	argv [i++] = filelist;
